import datetime

from flask_appbuilder._compat import as_unicode


def dict_to_json(xcol, ycols, labels, value_columns):  # pragma: no cover
    """
    Converts a list of dicts from datamodel query results
    to google chart json data.

    :param xcol:
        The name of a string column to be used has X axis on chart
    :param ycols:
        A list with the names of series cols, that can be used as numeric
    :param labels:
        A dict with the columns labels.
    :param value_columns:
        A list of dicts with the values to convert
    """
    json_data = dict()

    json_data["cols"] = [
        {"id": xcol, "label": as_unicode(labels[xcol]), "type": "string"}
    ]
    for ycol in ycols:
        json_data["cols"].append(
            {"id": ycol, "label": as_unicode(labels[ycol]), "type": "number"}
        )
    json_data["rows"] = []
    for value in value_columns:
        row = {"c": []}
        if isinstance(value[xcol], datetime.date):
            row["c"].append({"v": (str(value[xcol]))})
        else:
            row["c"].append({"v": (value[xcol])})
        for ycol in ycols:
            if value[ycol]:
                row["c"].append({"v": (value[ycol])})
            else:
                row["c"].append({"v": 0})
        json_data["rows"].append(row)
    return json_data



from datetime import datetime

from flask import g, request
from flask_appbuilder import ModelRestApi
from flask_appbuilder.api import expose, safe
from flask_appbuilder.const import API_RESULT_RES_KEY
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.security.decorators import permission_name, protect
from flask_appbuilder.security.sqla.apis.user.schema import (
    UserPostSchema,
    UserPutSchema,
)
from flask_appbuilder.security.sqla.models import Group, Role, User
from marshmallow import ValidationError
from sqlalchemy.exc import IntegrityError
from werkzeug.security import generate_password_hash


class UserApi(ModelRestApi):
    resource_name = "security/users"
    openapi_spec_tag = "Security Users"
    class_permission_name = "User"
    datamodel = SQLAInterface(User)
    allow_browser_login = True

    list_columns = [
        "id",
        "roles.id",
        "roles.name",
        "first_name",
        "last_name",
        "username",
        "active",
        "email",
        "last_login",
        "login_count",
        "fail_login_count",
        "created_on",
        "changed_on",
        "created_by.id",
        "changed_by.id",
        "groups",
    ]
    show_columns = list_columns
    add_columns = [
        "roles",
        "first_name",
        "last_name",
        "username",
        "active",
        "email",
        "password",
        "groups",
    ]
    edit_columns = add_columns
    search_columns = [
        "username",
        "first_name",
        "last_name",
        "active",
        "email",
        "created_by",
        "changed_by",
        "roles",
        "groups",
    ]

    add_model_schema = UserPostSchema()
    edit_model_schema = UserPutSchema()

    def pre_update(self, item, data):
        item.changed_on = datetime.now()
        item.changed_by_fk = g.user.id
        if "password" in data and data["password"]:
            item.password = generate_password_hash(
                password=data["password"],
                method=self.appbuilder.get_app.config.get(
                    "FAB_PASSWORD_HASH_METHOD", "scrypt"
                ),
                salt_length=self.appbuilder.get_app.config.get(
                    "FAB_PASSWORD_HASH_SALT_LENGTH", 16
                ),
            )

    def pre_add(self, item):
        item.password = generate_password_hash(
            password=item.password,
            method=self.appbuilder.get_app.config.get(
                "FAB_PASSWORD_HASH_METHOD", "scrypt"
            ),
            salt_length=self.appbuilder.get_app.config.get(
                "FAB_PASSWORD_HASH_SALT_LENGTH", 16
            ),
        )

    @expose("/", methods=["POST"])
    @protect()
    @safe
    @permission_name("post")
    def post(self):
        """Create new user
        ---
        post:
          requestBody:
            description: Model schema
            required: true
            content:
              application/json:
                schema:
                  $ref: '#/components/schemas/{{self.__class__.__name__}}.post'
          responses:
            201:
              description: Item changed
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      result:
                        $ref: '#/components/schemas/{{self.__class__.__name__}}.post'
            400:
              $ref: '#/components/responses/400'
            401:
              $ref: '#/components/responses/401'
            404:
              $ref: '#/components/responses/404'
            422:
              $ref: '#/components/responses/422'
            500:
              $ref: '#/components/responses/500'
        """
        try:
            item = self.add_model_schema.load(request.json)
            model = User()
            roles = []
            groups = []
            for key, value in item.items():
                if key not in ("roles", "groups"):
                    setattr(model, key, value)
                elif key == "roles":
                    roles = self._fetch_entities(Role, value)
                    missing_ids = set(value) - {r.id for r in roles}
                    if missing_ids:
                        return self.response_400(
                            message={
                                "roles": [
                                    (
                                        f"Role(s) with ID(s) {sorted(missing_ids)} "
                                        "not found."
                                    )
                                ]
                            }
                        )
                elif key == "groups":
                    groups = self._fetch_entities(Group, value)
                    missing_ids = set(value) - {g.id for g in groups}
                    if missing_ids:
                        return self.response_400(
                            message={
                                "groups": [
                                    (
                                        f"Group(s) with ID(s) {sorted(missing_ids)} "
                                        "not found."
                                    )
                                ]
                            }
                        )

            if "roles" in item.keys():
                model.roles = roles
            if "groups" in item.keys():
                model.groups = groups

            self.pre_add(model)
            self.datamodel.add(model, raise_exception=True)
            return self.response(201, id=model.id)
        except ValidationError as error:
            return self.response_400(message=error.messages)
        except IntegrityError as e:
            return self.response_422(message=str(e.orig))

    @expose("/<pk>", methods=["PUT"])
    @protect()
    @safe
    @permission_name("put")
    def put(self, pk):
        """Edit user
        ---
        put:
          parameters:
          - in: path
            schema:
              type: integer
            name: pk
          requestBody:
            description: Model schema
            required: true
            content:
              application/json:
                schema:
                  $ref: '#/components/schemas/{{self.__class__.__name__}}.put'
          responses:
            200:
              description: Item changed
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      result:
                        $ref: '#/components/schemas/{{self.__class__.__name__}}.put'
            400:
              $ref: '#/components/responses/400'
            401:
              $ref: '#/components/responses/401'
            404:
              $ref: '#/components/responses/404'
            422:
              $ref: '#/components/responses/422'
            500:
              $ref: '#/components/responses/500'
        """
        try:
            item = self.edit_model_schema.load(request.json)
            model = self.datamodel.get(pk, self._base_filters)
            roles = []
            groups = []

            item_roles = item.get("roles")
            item_groups = item.get("groups")

            if item_roles == [] and item_groups == []:
                return self.response_400(
                    message="User must have at least one role or group!"
                )

            if item_roles == [] and (item_groups is None and not model.groups):
                return self.response_400(
                    message=(
                        "Cannot clear all roles unless at least one group is \
                             assigned!"
                    )
                )

            if item_groups == [] and (item_roles is None and not model.roles):
                return self.response_400(
                    message=(
                        "Cannot clear all groups unless at least one role is \
                             assigned!"
                    )
                )

            for key, value in item.items():
                if key not in ("roles", "groups"):
                    setattr(model, key, value)
                elif key == "roles":
                    roles = self._fetch_entities(Role, value)
                    missing_ids = set(value) - {r.id for r in roles}
                    if missing_ids:
                        return self.response_404(
                            message={
                                "roles": [
                                    (
                                        f"Role(s) with ID(s) {sorted(missing_ids)} "
                                        "not found."
                                    )
                                ]
                            }
                        )
                elif key == "groups":
                    groups = self._fetch_entities(Group, value)
                    missing_ids = set(value) - {g.id for g in groups}
                    if missing_ids:
                        return self.response_404(
                            message={
                                "groups": [
                                    (
                                        f"Group(s) with ID(s) {sorted(missing_ids)} "
                                        "not found."
                                    )
                                ]
                            }
                        )

            if "roles" in item.keys():
                model.roles = roles
            if "groups" in item.keys():
                model.groups = groups

            self.pre_update(model, item)
            self.datamodel.edit(model, raise_exception=True)
            return self.response(
                200,
                **{API_RESULT_RES_KEY: self.edit_model_schema.dump(item, many=False)},
            )

        except ValidationError as e:
            return self.response_400(message=e.messages)
        except IntegrityError as e:
            return self.response_422(message=str(e.orig))



from flask_babel import lazy_gettext
from flask_wtf.recaptcha import RecaptchaField
from wtforms import BooleanField, PasswordField, StringField
from wtforms.validators import DataRequired, Email, EqualTo, ValidationError

from ..fieldwidgets import BS3PasswordFieldWidget, BS3TextFieldWidget
from ..forms import DynamicForm
from ..validators import PasswordComplexityValidator


class SelectDataRequired(DataRequired):
    """
    Select required flag on the input field will not work well on Chrome
    Console error:
        An invalid form control with name='roles' is not focusable.

    This makes a simple override to the DataRequired to be used specifically with
    select fields
    """

    field_flags = {}


def roles_or_groups_required(form, field):
    if not form["roles"].data and not form["groups"].data:
        raise ValidationError(lazy_gettext("Either select a role or a group"))


class LoginForm_oid(DynamicForm):
    openid = StringField(lazy_gettext("OpenID"), validators=[DataRequired()])
    username = StringField(lazy_gettext("User Name"))
    remember_me = BooleanField(lazy_gettext("Remember me"), default=False)


class LoginForm_db(DynamicForm):
    username = StringField(lazy_gettext("User Name"), validators=[DataRequired()])
    password = PasswordField(lazy_gettext("Password"), validators=[DataRequired()])


class UserInfoEdit(DynamicForm):
    first_name = StringField(
        lazy_gettext("First Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
        description=lazy_gettext("Write the user first name or names"),
    )
    last_name = StringField(
        lazy_gettext("Last Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
        description=lazy_gettext("Write the user last name"),
    )


class ResetPasswordForm(DynamicForm):
    password = PasswordField(
        lazy_gettext("Password"),
        description=lazy_gettext(
            "Please use a good password policy,"
            " this application does not check this for you"
        ),
        validators=[DataRequired(), PasswordComplexityValidator()],
        widget=BS3PasswordFieldWidget(),
    )
    conf_password = PasswordField(
        lazy_gettext("Confirm Password"),
        description=lazy_gettext("Please rewrite the password to confirm"),
        validators=[EqualTo("password", message=lazy_gettext("Passwords must match"))],
        widget=BS3PasswordFieldWidget(),
    )


class RegisterUserDBForm(DynamicForm):
    username = StringField(
        lazy_gettext("User Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    first_name = StringField(
        lazy_gettext("First Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    last_name = StringField(
        lazy_gettext("Last Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    email = StringField(
        lazy_gettext("Email"),
        validators=[DataRequired(), Email()],
        widget=BS3TextFieldWidget(),
    )
    password = PasswordField(
        lazy_gettext("Password"),
        description=lazy_gettext(
            "Please use a good password policy,"
            " this application does not check this for you"
        ),
        validators=[DataRequired()],
        widget=BS3PasswordFieldWidget(),
    )
    conf_password = PasswordField(
        lazy_gettext("Confirm Password"),
        description=lazy_gettext("Please rewrite the password to confirm"),
        validators=[EqualTo("password", message=lazy_gettext("Passwords must match"))],
        widget=BS3PasswordFieldWidget(),
    )
    recaptcha = RecaptchaField()


class RegisterUserOIDForm(DynamicForm):
    username = StringField(
        lazy_gettext("User Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    first_name = StringField(
        lazy_gettext("First Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    last_name = StringField(
        lazy_gettext("Last Name"),
        validators=[DataRequired()],
        widget=BS3TextFieldWidget(),
    )
    email = StringField(
        lazy_gettext("Email"),
        validators=[DataRequired(), Email()],
        widget=BS3TextFieldWidget(),
    )
    recaptcha = RecaptchaField()



from flask_appbuilder import ModelRestApi, BaseView, has_access, expose
from flask_appbuilder.models.sqla.interface import SQLAInterface

from . import appbuilder, db
from .models import Contact, ContactGroup, Gender


def fill_gender():
    try:
        db.session.add(Gender(name="Male"))
        db.session.add(Gender(name="Female"))
        db.session.commit()
    except Exception:
        db.session.rollback()


db.create_all()
fill_gender()


class ContactModelApi(ModelRestApi):
    resource_name = "contact"
    datamodel = SQLAInterface(Contact)
    allow_browser_login = True
    list_columns = ["name", "address", 'birthday', 'contact_group.name']
    show_columns = [
            "name",
            "address",
            'birthday',
            'contact_group.name',
            "personal_celphone",
            "personal_phone",
            "gender.name"]


appbuilder.add_api(ContactModelApi)


class GroupModelApi(ModelRestApi):
   resource_name = "group"
   datamodel = SQLAInterface(ContactGroup)
   allow_browser_login = True


appbuilder.add_api(GroupModelApi)


class ReactRenderView(BaseView):
    @expose('/<string:param1>')
    @has_access
    def render_react(self, param1):
        self.update_redirect()
        return self.render_template('react.html',
                                    param1 = param1)


appbuilder.add_view_no_menu(ReactRenderView)
appbuilder.add_link("Contacts", href='/reactrenderview/contact', category='Contacts')
appbuilder.add_link("Groups", href='/reactrenderview/group', category='Contacts')




import calendar

from flask import make_response, Response
from flask_appbuilder import expose, has_access, permission_name
from flask_appbuilder import ModelView
from flask_appbuilder.charts.views import GroupByChartView
from flask_appbuilder.models.group import aggregate_count
from flask_appbuilder.models.mongoengine.interface import MongoEngineInterface


from . import appbuilder
from .models import ContactGroup, Contact, Tags, Gender


def fill_gender():
    try:
        g1 = Gender(name="Male")
        g1.save()
        g2 = Gender(name="Female")
        g2.save()
    except:
        pass


class ContactModelView(ModelView):
    datamodel = MongoEngineInterface(Contact)
    label_columns = {"image_thumb_show": "Photo", "image_show": "Photo"}
    list_columns = [
        "image_thumb_show",
        "name",
        "personal_celphone",
        "birthday",
        "contact_group",
    ]
    show_columns = [
        "image_show",
        "name",
        "personal_celphone",
        "birthday",
        "contact_group",
    ]

    @expose("/mongo_download/<pk>")
    @has_access
    def mongo_download(self, pk):
        item = self.datamodel.get(pk)
        file = item.file.read()
        response = make_response(file)
        response.headers["Content-Disposition"] = "attachment; filename={0}".format(
            item.file.name
        )
        return response

    @expose("/img/<pk>")
    @has_access
    @permission_name("show_img")
    def img(self, pk):
        item = self.datamodel.get(pk)
        mime_type = item.image.content_type
        return Response(item.image.read(), mimetype=mime_type, direct_passthrough=True)

    @expose("/img_thumb/<pk>")
    @has_access
    @permission_name("show_img")
    def img_thumb(self, pk):
        item = self.datamodel.get(pk)
        mime_type = item.image.content_type
        return Response(
            item.image.thumbnail.read(), mimetype=mime_type, direct_passthrough=True
        )


class GroupModelView(ModelView):
    datamodel = MongoEngineInterface(ContactGroup)
    related_views = [ContactModelView]
    search_columns = ["name"]


class TagsModelView(ModelView):
    datamodel = MongoEngineInterface(Tags)


class ContactChartView(GroupByChartView):
    datamodel = MongoEngineInterface(Contact)
    chart_title = "Grouped contacts"
    label_columns = ContactModelView.label_columns
    chart_type = "PieChart"

    definitions = [
        {"group": "contact_group", "series": [(aggregate_count, "contact_group")]},
        {"group": "gender", "series": [(aggregate_count, "gender")]},
    ]


def pretty_month_year(value):
    return calendar.month_name[value.month] + " " + str(value.year)


def pretty_year(value):
    return str(value.year)


class ContactTimeChartView(GroupByChartView):
    datamodel = MongoEngineInterface(Contact)

    chart_title = "Grouped Birth contacts"
    chart_type = "AreaChart"
    label_columns = ContactModelView.label_columns
    definitions = [
        {
            "group": "month_year",
            "formatter": pretty_month_year,
            "series": [(aggregate_count, "contact_group")],
        },
        {
            "group": "year",
            "formatter": pretty_year,
            "series": [(aggregate_count, "contact_group")],
        },
    ]


appbuilder.add_view(
    GroupModelView,
    "List Groups",
    icon="fa-folder-open-o",
    category="Contacts",
    category_icon="fa-envelope",
)
appbuilder.add_view(
    ContactModelView,
    "List Contacts",
    icon="fa-folder-open-o",
    category="Contacts",
    category_icon="fa-envelope",
)
appbuilder.add_view(
    TagsModelView,
    "List Tags",
    icon="fa-folder-open-o",
    category="Contacts",
    category_icon="fa-envelope",
)
appbuilder.add_separator("Contacts")
appbuilder.add_view(
    ContactChartView, "Contacts Chart", icon="fa-dashboard", category="Contacts"
)
appbuilder.add_view(
    ContactTimeChartView,
    "Contacts Birth Chart",
    icon="fa-dashboard",
    category="Contacts",
)

appbuilder.security_cleanup()
fill_gender()



from datetime import datetime
import logging
import random

from app import db
from app.models import Contact, ContactGroup, Gender, ModelOMParent, ModelOMChild

log = logging.getLogger(__name__)


def get_random_name(names_list, size=1):
    name_lst = [
        names_list[random.randrange(0, len(names_list))].decode("utf-8").capitalize()
        for i in range(0, size)
    ]
    return " ".join(name_lst)


try:
    db.session.query(Contact).delete()
    db.session.query(ContactGroup).delete()
    db.session.commit()
except Exception:
    db.session.rollback()

try:
    groups = []
    groups.append(ContactGroup(name="Friends"))
    groups.append(ContactGroup(name="Family"))
    groups.append(ContactGroup(name="Work"))
    db.session.add(groups[0])
    db.session.add(groups[1])
    db.session.add(groups[2])
    print(groups[0].id)
    db.session.commit()
except Exception as e:
    log.error("Creating Groups: %s", e)
    db.session.rollback()


model_oo_parents = list()
for i in range(20):
    model = ModelOMParent()
    model.field_string = f"text{i}"
    db.session.add(model)
    db.session.commit()
    model_oo_parents.append(model)

for i in range(20):
    for j in range(1, 4):
        model = ModelOMChild()
        model.field_string = f"text{i}.{j}"
        model.parent = model_oo_parents[i]
        db.session.add(model)
        db.session.commit()


f = open("NAMES.DIC", "rb")
names_list = [x.strip() for x in f.readlines()]

f.close()

for i in range(1, 1000):
    c = Contact()
    c.name = get_random_name(names_list, random.randrange(2, 6))
    c.address = "Street " + names_list[random.randrange(0, len(names_list))].decode(
        "utf-8"
    )
    c.personal_phone = random.randrange(1111111, 9999999)
    c.personal_celphone = random.randrange(1111111, 9999999)
    c.contact_group = groups[random.randrange(0, 3)]
    c.gender = random.choice(list(Gender))
    year = random.choice(range(1900, 2012))
    month = random.choice(range(1, 12))
    day = random.choice(range(1, 28))
    c.birthday = datetime(year, month, day)
    db.session.add(c)
    try:
        db.session.commit()
        print("inserted", c)
    except Exception as e:
        log.error("Creating Contact: %s", e)
        db.session.rollback()



from flask_appbuilder.baseviews import BaseView, expose

from . import appbuilder


class MyView(BaseView):
    route_base = "/myview"

    @expose("/method1/<string:param1>")
    def method1(self, param1):
        # do something with param1
        # and return to previous page or index
        param1 = "Hello %s" % (param1)
        return param1

    @expose("/method2/<string:param1>")
    def method2(self, param1):
        # do something with param1
        # and render template with param
        param1 = "Goodbye %s" % (param1)
        return param1


appbuilder.add_view_no_menu(MyView())



import json
import logging
import os.path as op
from typing import Set

from flask import (
    abort,
    flash,
    jsonify,
    make_response,
    redirect,
    request,
    send_file,
    session,
    url_for,
)
from flask_appbuilder.exceptions import FABException


from ._compat import as_unicode, string_types
from .baseviews import BaseCRUDView, BaseFormView, BaseView, expose, expose_api
from .const import FLAMSG_ERR_SEC_ACCESS_DENIED, PERMISSION_PREFIX
from .filemanager import uuid_originalname
from .security.decorators import has_access, has_access_api, permission_name
from .urltools import get_filter_args, get_order_args, get_page_args, get_page_size_args
from .widgets import GroupFormListWidget, ListMasterWidget

log = logging.getLogger(__name__)


class IndexView(BaseView):
    """
    A simple view that implements the index for the site
    """

    route_base = ""
    default_view = "index"
    index_template = "appbuilder/index.html"

    @expose("/")
    def index(self):
        self.update_redirect()
        return self.render_template(self.index_template, appbuilder=self.appbuilder)


class UtilView(BaseView):
    """
    A simple view that implements special util routes.
    At the moment it only supports the back special endpoint.
    """

    route_base = ""
    default_view = "back"

    @expose("/back")
    def back(self):
        return redirect(self.get_redirect())


class SimpleFormView(BaseFormView):
    """
    View for presenting your own forms
    Inherit from this view to provide some base processing
    for your customized form views.

    Notice that this class inherits from BaseView so all properties
    from the parent class can be overridden also.

    Implement form_get and form_post to implement your
    form pre-processing and post-processing
    """

    @expose("/form", methods=["GET"])
    @has_access
    def this_form_get(self):
        self._init_vars()
        form = self.form.refresh()

        self.form_get(form)
        widgets = self._get_edit_widget(form=form)
        self.update_redirect()
        return self.render_template(
            self.form_template,
            title=self.form_title,
            widgets=widgets,
            appbuilder=self.appbuilder,
        )

    @expose("/form", methods=["POST"])
    @has_access
    def this_form_post(self):
        self._init_vars()
        form = self.form.refresh()

        if form.validate_on_submit():
            response = self.form_post(form)
            if not response:
                return redirect(self.get_redirect())
            return response
        else:
            widgets = self._get_edit_widget(form=form)
            return self.render_template(
                self.form_template,
                title=self.form_title,
                widgets=widgets,
                appbuilder=self.appbuilder,
            )


class PublicFormView(BaseFormView):
    """
    View for presenting your own forms
    Inherit from this view to provide some base
    processing for your customized form views.

    Notice that this class inherits from BaseView
    so all properties from the parent class can be overridden also.

    Implement form_get and form_post to implement
    your form pre-processing and post-processing
    """

    @expose("/form", methods=["GET"])
    def this_form_get(self):
        self._init_vars()
        form = self.form.refresh()
        self.form_get(form)
        widgets = self._get_edit_widget(form=form)
        self.update_redirect()
        return self.render_template(
            self.form_template,
            title=self.form_title,
            widgets=widgets,
            appbuilder=self.appbuilder,
        )

    @expose("/form", methods=["POST"])
    def this_form_post(self):
        self._init_vars()
        form = self.form.refresh()
        if form.validate_on_submit():
            response = self.form_post(form)
            if not response:
                return redirect(self.get_redirect())
            return response
        else:
            widgets = self._get_edit_widget(form=form)
            return self.render_template(
                self.form_template,
                title=self.form_title,
                widgets=widgets,
                appbuilder=self.appbuilder,
            )


class RestCRUDView(BaseCRUDView):
    """
    This class view exposes REST method for CRUD operations on you models
    """

    disable_api_route_methods: bool = False
    """ Flag to disable this class exposed methods, note that this class
    will eventually get deprecated """

    def __init__(self, **kwargs):
        if self.disable_api_route_methods:
            api_route_methods: Set = {
                "api",
                "api_read",
                "api_get",
                "api_create",
                "api_update",
                "api_delete",
                "api_column_add",
                "api_column_edit",
                "api_readvalues",
            }
            self.exclude_route_methods = self.exclude_route_methods | api_route_methods
        super().__init__(**kwargs)

    def _search_form_json(self):
        pass

    def _get_api_urls(self, api_urls=None):
        """
            Completes a dict with the CRUD urls of the API.

        :param api_urls: A dict with the urls {'<FUNCTION>':'<URL>',...}
        :return: A dict with the CRUD urls of the base API.
        """
        view_name = self.__class__.__name__
        api_urls = api_urls or {}
        api_urls["read"] = url_for(view_name + ".api_read")
        api_urls["delete"] = url_for(view_name + ".api_delete", pk="")
        api_urls["create"] = url_for(view_name + ".api_create")
        api_urls["update"] = url_for(view_name + ".api_update", pk="")
        return api_urls

    def _get_modelview_urls(self, modelview_urls=None):
        view_name = self.__class__.__name__
        modelview_urls = modelview_urls or {}
        modelview_urls["show"] = url_for(view_name + ".show", pk="")
        modelview_urls["add"] = url_for(view_name + ".add")
        modelview_urls["edit"] = url_for(view_name + ".edit", pk="")
        return modelview_urls

    @expose("/api", methods=["GET"])
    @has_access_api
    @permission_name("list")
    def api(self):
        log.warning("This API is deprecated and will be removed on 1.15.X")
        view_name = self.__class__.__name__
        api_urls = self._get_api_urls()
        modelview_urls = self._get_modelview_urls()
        #
        # Collects the CRUD permissions
        can_show = self.appbuilder.sm.has_access("can_show", view_name)
        can_edit = self.appbuilder.sm.has_access("can_edit", view_name)
        can_add = self.appbuilder.sm.has_access("can_add", view_name)
        can_delete = self.appbuilder.sm.has_access("can_delete", view_name)
        #
        # Prepares the form with the search fields make it JSON serializable
        form_fields = {}
        search_filters = {}
        dict_filters = self._filters.get_search_filters()
        form = self.search_form.refresh()
        for col in self.search_columns:
            form_fields[col] = form[col]()
            search_filters[col] = [as_unicode(flt.name) for flt in dict_filters[col]]

        ret_json = jsonify(
            can_show=can_show,
            can_add=can_add,
            can_edit=can_edit,
            can_delete=can_delete,
            label_columns=self._label_columns_json(),
            list_columns=self.list_columns,
            order_columns=self.order_columns,
            page_size=self.page_size,
            modelview_name=view_name,
            api_urls=api_urls,
            search_filters=search_filters,
            search_fields=form_fields,
            modelview_urls=modelview_urls,
        )
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response

    @expose_api(name="read", url="/api/read", methods=["GET"])
    @has_access_api
    @permission_name("list")
    def api_read(self):
        """ """
        log.warning("This API is deprecated and will be removed on 2.3.X")
        # Get arguments for ordering
        if get_order_args().get(self.__class__.__name__):
            order_column, order_direction = get_order_args().get(
                self.__class__.__name__
            )
        else:
            order_column, order_direction = "", ""
        page = get_page_args().get(self.__class__.__name__)
        page_size = get_page_size_args().get(self.__class__.__name__)
        get_filter_args(self._filters)
        joined_filters = self._filters.get_joined_filters(self._base_filters)
        count, lst = self.datamodel.query(
            joined_filters,
            order_column,
            order_direction,
            page=page,
            page_size=page_size,
        )
        result = self.datamodel.get_values_json(lst, self.list_columns)
        pks = self.datamodel.get_keys(lst)
        ret_json = jsonify(
            label_columns=self._label_columns_json(),
            list_columns=self.list_columns,
            order_columns=self.order_columns,
            page=page,
            page_size=page_size,
            count=count,
            modelview_name=self.__class__.__name__,
            pks=pks,
            result=result,
        )
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response

    def show_item_dict(self, item):
        """Returns a json-able dict for show"""
        d = {}
        for col in self.show_columns:
            v = getattr(item, col)
            if not isinstance(v, (int, float, string_types)):
                v = str(v)
            d[col] = v
        return d

    @expose_api(name="get", url="/api/get/<pk>", methods=["GET"])
    @has_access_api
    @permission_name("show")
    def api_get(self, pk):
        """ """
        log.warning("This API is deprecated and will be removed on 2.3.X")
        # Get arguments for ordering
        item = self.datamodel.get(pk, self._base_filters)
        if not item:
            abort(404)
        ret_json = jsonify(
            pk=pk,
            label_columns=self._label_columns_json(),
            include_columns=self.show_columns,
            modelview_name=self.__class__.__name__,
            result=self.show_item_dict(item),
        )
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response

    @expose_api(name="create", url="/api/create", methods=["POST"])
    @has_access_api
    @permission_name("add")
    def api_create(self):
        log.warning("This API is deprecated and will be removed on 2.3.X")
        get_filter_args(self._filters, disallow_if_not_in_search=False)
        exclude_cols = self._filters.get_relation_cols()
        form = self.add_form.refresh()
        self._fill_form_exclude_cols(exclude_cols, form)
        if form.validate():
            item = self.datamodel.obj()
            form.populate_obj(item)
            self.pre_add(item)
            if self.datamodel.add(item):
                self.post_add(item)
                http_return_code = 200
            else:
                http_return_code = 500
            payload = {
                "message": self.datamodel.message[0],
                "item": self.show_item_dict(item),
                "severity": self.datamodel.message[1],
            }
        else:
            payload = {"message": "Validation error", "error_details": form.errors}
            http_return_code = 500
        return make_response(jsonify(payload), http_return_code)

    @expose_api(name="update", url="/api/update/<pk>", methods=["PUT"])
    @has_access_api
    @permission_name("edit")
    def api_update(self, pk):
        log.warning("This API is deprecated and will be removed on 2.3.X")
        get_filter_args(self._filters, disallow_if_not_in_search=False)
        exclude_cols = self._filters.get_relation_cols()

        item = self.datamodel.get(pk, self._base_filters)
        if not item:
            abort(404)
        # convert pk to correct type, if pk is non string type.
        pk = self.datamodel.get_pk_value(item)

        form = self.edit_form.refresh(request.form)
        # fill the form with the suppressed cols, generated from exclude_cols
        self._fill_form_exclude_cols(exclude_cols, form)
        # trick to pass unique validation
        form._id = pk
        http_return_code = 500
        if form.validate():
            # Deleting form fields not specified as keys in POST data
            # this allows for other Model columns to be left untouched when
            # unspecified.
            form_fields = set([t for t in form._fields.keys()])
            for field in form_fields - set(request.form.keys()):
                delattr(form, field)

            form.populate_obj(item)
            self.pre_update(item)
            if self.datamodel.edit(item):
                self.post_update(item)
                http_return_code = 200
            payload = {
                "message": self.datamodel.message[0],
                "severity": self.datamodel.message[1],
                "item": self.show_item_dict(item),
            }
        else:
            payload = {
                "message": "Validation error",
                "error_details": form.errors,
                "severity": "warning",
            }
        return make_response(jsonify(payload), http_return_code)

    @expose_api(name="delete", url="/api/delete/<pk>", methods=["DELETE"])
    @has_access_api
    @permission_name("delete")
    def api_delete(self, pk):
        log.warning("This API is deprecated and will be removed on 2.3.X")
        item = self.datamodel.get(pk, self._base_filters)
        if not item:
            abort(404)
        self.pre_delete(item)
        if self.datamodel.delete(item):
            self.post_delete(item)
            http_return_code = 200
        else:
            http_return_code = 500
        response = make_response(
            jsonify(
                {
                    "message": self.datamodel.message[0],
                    "severity": self.datamodel.message[1],
                }
            ),
            http_return_code,
        )
        response.headers["Content-Type"] = "application/json"
        return response

    def _get_related_column_data(self, col_name, filters):
        rel_datamodel = self.datamodel.get_related_interface(col_name)
        _filters = rel_datamodel.get_filters(rel_datamodel.get_search_columns_list())
        get_filter_args(_filters)
        if filters:
            filters = _filters.add_filter_list(filters)
        else:
            filters = _filters
        result = rel_datamodel.query(filters)[1]
        ret_list = list()
        for item in result:
            pk = rel_datamodel.get_pk_value(item)
            ret_list.append({"id": int(pk), "text": str(item)})
        ret_json = json.dumps(ret_list)
        return ret_json

    @expose_api(name="column_add", url="/api/column/add/<col_name>", methods=["GET"])
    @has_access_api
    @permission_name("add")
    def api_column_add(self, col_name):
        """
            Returns list of (pk, object) nice to use on select2.
            Use only for related columns.
            Always filters with add_form_query_rel_fields, and accepts extra filters
            on endpoint arguments.
        :param col_name: The related column name
        :return: JSON response
        """
        log.warning("This API is deprecated and will be removed on 2.3.X")
        filter_rel_fields = None
        if self.add_form_query_rel_fields:
            filter_rel_fields = self.add_form_query_rel_fields.get(col_name)
        ret_json = self._get_related_column_data(col_name, filter_rel_fields)
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response

    @expose_api(name="column_edit", url="/api/column/edit/<col_name>", methods=["GET"])
    @has_access_api
    @permission_name("edit")
    def api_column_edit(self, col_name):
        """
            Returns list of (pk, object) nice to use on select2.
            Use only for related columns.
            Always filters with edit_form_query_rel_fields, and accepts extra filters
            on endpoint arguments.
        :param col_name: The related column name
        :return: JSON response
        """
        log.warning("This API is deprecated and will be removed on 2.3.X")
        filter_rel_fields = None
        if self.edit_form_query_rel_fields:
            filter_rel_fields = self.edit_form_query_rel_fields.get(col_name)
        ret_json = self._get_related_column_data(col_name, filter_rel_fields)
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response

    @expose_api(name="readvalues", url="/api/readvalues", methods=["GET"])
    @has_access_api
    @permission_name("list")
    def api_readvalues(self):
        """ """
        log.warning("This API is deprecated and will be removed on 2.3.X")
        # Get arguments for ordering
        if get_order_args().get(self.__class__.__name__):
            order_column, order_direction = get_order_args().get(
                self.__class__.__name__
            )
        else:
            order_column, order_direction = "", ""
        get_filter_args(self._filters)
        joined_filters = self._filters.get_joined_filters(self._base_filters)
        count, result = self.datamodel.query(
            joined_filters, order_column, order_direction
        )

        ret_list = list()
        for item in result:
            pk = self.datamodel.get_pk_value(item)
            ret_list.append({"id": int(pk), "text": str(item)})

        ret_json = json.dumps(ret_list)
        response = make_response(ret_json, 200)
        response.headers["Content-Type"] = "application/json"
        return response


class ModelView(RestCRUDView):
    """
    This is the CRUD generic view.
    If you want to automatically implement create, edit,
    delete, show, and list from your database tables,
    inherit your views from this class.

    Notice that this class inherits from BaseCRUDView and BaseModelView
    so all properties from the parent class can be overridden.
    """

    def __init__(self, **kwargs):
        super(ModelView, self).__init__(**kwargs)

    def post_add_redirect(self):
        """Override this function to control the
        redirect after add endpoint is called."""
        return redirect(self.get_redirect())

    def post_edit_redirect(self):
        """Override this function to control the
        redirect after edit endpoint is called."""
        return redirect(self.get_redirect())

    def post_delete_redirect(self):
        """Override this function to control the
        redirect after edit endpoint is called."""
        return redirect(self.get_redirect())

    """
    --------------------------------
            LIST
    --------------------------------
    """

    @expose("/list/")
    @has_access
    def list(self):
        self.update_redirect()
        try:
            widgets = self._list()
        except FABException as exc:
            flash(f"An error occurred: {exc}", "warning")
            return redirect(self.get_redirect())
        return self.render_template(
            self.list_template, title=self.list_title, widgets=widgets
        )

    """
    --------------------------------
            SHOW
    --------------------------------
    """

    @expose("/show/<pk>", methods=["GET"])
    @has_access
    def show(self, pk):
        pk = self._deserialize_pk_if_composite(pk)
        widgets = self._show(pk)
        return self.render_template(
            self.show_template,
            pk=pk,
            title=self.show_title,
            widgets=widgets,
            related_views=self._related_views,
        )

    """
    ---------------------------
            ADD
    ---------------------------
    """

    @expose("/add", methods=["GET", "POST"])
    @has_access
    def add(self):
        widget = self._add()
        if not widget:
            return self.post_add_redirect()
        else:
            return self.render_template(
                self.add_template, title=self.add_title, widgets=widget
            )

    """
    ---------------------------
            EDIT
    ---------------------------
    """

    @expose("/edit/<pk>", methods=["GET", "POST"])
    @has_access
    def edit(self, pk):
        pk = self._deserialize_pk_if_composite(pk)
        widgets = self._edit(pk)
        if not widgets:
            return self.post_edit_redirect()
        else:
            return self.render_template(
                self.edit_template,
                title=self.edit_title,
                widgets=widgets,
                related_views=self._related_views,
            )

    """
    ---------------------------
            DELETE
    ---------------------------
    """

    @expose("/delete/<pk>", methods=["GET", "POST"])
    @has_access
    def delete(self, pk):
        # Maintains compatibility but refuses to delete on GET methods if CSRF is enabled
        if not self.is_get_mutation_allowed():
            self.update_redirect()
            log.warning("CSRF is enabled and a delete using GET was invoked")
            flash(as_unicode(FLAMSG_ERR_SEC_ACCESS_DENIED), "danger")
            return self.post_delete_redirect()
        pk = self._deserialize_pk_if_composite(pk)
        self._delete(pk)
        return self.post_delete_redirect()

    @expose("/download/<string:filename>")
    @has_access
    def download(self, filename):
        return send_file(
            op.join(self.appbuilder.app.config["UPLOAD_FOLDER"], filename),
            download_name=uuid_originalname(filename),
            as_attachment=True,
        )

    def get_action_permission_name(self, name: str) -> str:
        """
        Get the permission name of an action name
        """
        _permission_name = self.method_permission_name.get(
            self.actions.get(name).func.__name__
        )
        if _permission_name:
            return PERMISSION_PREFIX + _permission_name
        else:
            return name

    @expose("/action/<string:name>/<pk>", methods=["GET", "POST"])
    def action(self, name, pk):
        """
        Action method to handle actions from a show view
        """
        # Maintains compatibility but refuses to proceed if CSRF is enabled
        if not self.is_get_mutation_allowed():
            self.update_redirect()
            log.warning("CSRF is enabled and a action using GET was invoked")
            flash(as_unicode(FLAMSG_ERR_SEC_ACCESS_DENIED), "danger")
            return redirect(self.get_redirect())

        pk = self._deserialize_pk_if_composite(pk)
        permission_name = self.get_action_permission_name(name)
        if self.appbuilder.sm.has_access(permission_name, self.class_permission_name):
            action = self.actions.get(name)
            return action.func(self.datamodel.get(pk))
        else:
            flash(as_unicode(FLAMSG_ERR_SEC_ACCESS_DENIED), "danger")
            return redirect(".")

    @expose("/action_post", methods=["POST"])
    def action_post(self):
        """
        Action method to handle multiple records selected from a list view
        """
        name = request.form["action"]
        pks = request.form.getlist("rowid")
        permission_name = self.get_action_permission_name(name)

        if self.appbuilder.sm.has_access(permission_name, self.class_permission_name):
            action = self.actions.get(name)
            items = [
                self.datamodel.get(self._deserialize_pk_if_composite(pk)) for pk in pks
            ]
            return action.func(items)
        else:
            flash(as_unicode(FLAMSG_ERR_SEC_ACCESS_DENIED), "danger")
            return redirect(".")


class MasterDetailView(BaseCRUDView):
    """
    Implements behaviour for controlling two CRUD views
    linked by PK and FK, in a master/detail type with
    two lists.

    Master view will behave like a left menu::

        class DetailView(ModelView):
            datamodel = SQLAInterface(DetailTable, db.session)

        class MasterView(MasterDetailView):
            datamodel = SQLAInterface(MasterTable, db.session)
            related_views = [DetailView]

    """

    list_template = "appbuilder/general/model/left_master_detail.html"
    list_widget = ListMasterWidget
    master_div_width = 2
    """
        Set to configure bootstrap class for master grid size
    """

    @expose("/list/")
    @expose("/list/<pk>")
    @has_access
    def list(self, pk=None):
        pages = get_page_args()
        page_sizes = get_page_size_args()
        orders = get_order_args()

        widgets = self._list()
        if pk:
            item = self.datamodel.get(pk)
            widgets = self._get_related_views_widgets(
                item, orders=orders, pages=pages, page_sizes=page_sizes, widgets=widgets
            )
            related_views = self._related_views
        else:
            related_views = []

        return self.render_template(
            self.list_template,
            title=self.list_title,
            widgets=widgets,
            related_views=related_views,
            master_div_width=self.master_div_width,
        )


class MultipleView(BaseView):
    """
    Use this view to render multiple views on the same page,
    exposed on the list endpoint.

    example (after defining GroupModelView and ContactModelView)::

        class MultipleViewsExp(MultipleView):
            views = [GroupModelView, ContactModelView]

    """

    list_template = "appbuilder/general/model/multiple_views.html"
    """ Override this to implement your own template for the list endpoint """

    views = None
    " A list of ModelView's to render on the same page "
    _views = None

    def __init__(self, **kwargs):
        super(MultipleView, self).__init__(**kwargs)
        self.views = self.views or list()
        self._views = self._views or list()

    def get_uninit_inner_views(self):
        return self.views

    def get_init_inner_views(self):
        return self._views

    @expose("/list/")
    @has_access
    def list(self):
        pages = get_page_args()
        page_sizes = get_page_size_args()
        orders = get_order_args()
        views_widgets = list()
        for view in self._views:
            if orders.get(view.__class__.__name__):
                order_column, order_direction = orders.get(view.__class__.__name__)
            else:
                order_column, order_direction = "", ""
            page = pages.get(view.__class__.__name__)
            page_size = page_sizes.get(view.__class__.__name__)
            views_widgets.append(
                view._get_view_widget(
                    filters=view._base_filters,
                    order_column=order_column,
                    order_direction=order_direction,
                    page=page,
                    page_size=page_size,
                )
            )
        self.update_redirect()
        return self.render_template(
            self.list_template, views=self._views, views_widgets=views_widgets
        )


class CompactCRUDMixin(BaseCRUDView):
    """
    Mix with ModelView to implement a list with add and edit on the same page.
    """

    @classmethod
    def set_key(cls, k, v):
        """Allows attaching stateless information to the class using the
        flask session dict
        """
        k = cls.__name__ + "__" + k
        session[k] = v

    @classmethod
    def get_key(cls, k, default=None):
        """Matching get method for ``set_key``"""
        k = cls.__name__ + "__" + k
        if k in session:
            return session[k]
        else:
            return default

    @classmethod
    def del_key(cls, k):
        """Matching get method for ``set_key``"""
        k = cls.__name__ + "__" + k
        session.pop(k)

    def _get_list_widget(self, **kwargs):
        """get joined base filter and current active filter for query"""
        widgets = super(CompactCRUDMixin, self)._get_list_widget(**kwargs)
        session_form_widget = self.get_key("session_form_widget", None)

        form_widget = None
        if session_form_widget == "add":
            form_widget = self._add().get("add")
        elif session_form_widget == "edit":
            pk = self.get_key("session_form_edit_pk")
            if pk and self.datamodel.get(int(pk)):
                form_widget = self._edit(int(pk)).get("edit")
        return {
            "list": GroupFormListWidget(
                list_widget=widgets.get("list"),
                form_widget=form_widget,
                form_action=self.get_key("session_form_action", ""),
                form_title=self.get_key("session_form_title", ""),
            )
        }

    @expose("/list/", methods=["GET", "POST"])
    @has_access
    def list(self):
        list_widgets = self._list()
        return self.render_template(
            self.list_template, title=self.list_title, widgets=list_widgets
        )

    @expose("/add/", methods=["GET", "POST"])
    @has_access
    def add(self):
        widgets = self._add()
        if not widgets:
            self.set_key("session_form_action", "")
            self.set_key("session_form_widget", None)
            return redirect(request.referrer)
        else:
            self.set_key("session_form_widget", "add")
            self.set_key("session_form_action", request.script_root + request.full_path)
            self.set_key("session_form_title", self.add_title)
            return redirect(self.get_redirect())

    @expose("/edit/<pk>", methods=["GET", "POST"])
    @has_access
    def edit(self, pk):
        pk = self._deserialize_pk_if_composite(pk)
        widgets = self._edit(pk)
        self.update_redirect()
        if not widgets:
            self.set_key("session_form_action", "")
            self.set_key("session_form_widget", None)
            return redirect(self.get_redirect())
        else:
            self.set_key("session_form_widget", "edit")
            self.set_key("session_form_action", request.script_root + request.full_path)
            self.set_key("session_form_title", self.add_title)
            self.set_key("session_form_edit_pk", pk)
            return redirect(self.get_redirect())

    @expose("/delete/<pk>", methods=["GET", "POST"])
    @has_access
    def delete(self, pk):
        # Maintains compatibility but refuses to delete on GET methods if CSRF is enabled
        if not self.is_get_mutation_allowed():
            self.update_redirect()
            log.warning("CSRF is enabled and a delete using GET was invoked")
            flash(as_unicode(FLAMSG_ERR_SEC_ACCESS_DENIED), "danger")
            return redirect(self.get_redirect())

        pk = self._deserialize_pk_if_composite(pk)
        self._delete(pk)
        edit_pk = self.get_key("session_form_edit_pk")
        if pk == edit_pk:
            self.del_key("session_form_edit_pk")
        return redirect(self.get_redirect())


"""
    This is for retro compatibility
"""
GeneralView = ModelView



import os

from flask_appbuilder.const import AUTH_DB
from flask_appbuilder.exceptions import PasswordComplexityValidationError

basedir = os.path.abspath(os.path.dirname(__file__))

CSRF_ENABLED = True
SECRET_KEY = "\2\1thisismyscretkey\1\2\e\y\y\h"

OPENID_PROVIDERS = [
    {"name": "Google", "url": "https://www.google.com/accounts/o8/id"},
    {"name": "Yahoo", "url": "https://me.yahoo.com"},
    {"name": "AOL", "url": "http://openid.aol.com/<username>"},
    {"name": "Flickr", "url": "http://www.flickr.com/<username>"},
    {"name": "MyOpenID", "url": "https://www.myopenid.com"},
]

SQLALCHEMY_DATABASE_URI = "sqlite:///" + os.path.join(basedir, "app.db")
# SQLALCHEMY_DATABASE_URI = 'mysql://username:password@mysqlserver.local/quickhowto'
# SQLALCHEMY_DATABASE_URI = 'postgresql://scott:tiger@localhost:5432/myapp'
# SQLALCHEMY_ECHO = True
SQLALCHEMY_POOL_RECYCLE = 3

BABEL_DEFAULT_LOCALE = "en"
BABEL_DEFAULT_FOLDER = "translations"
LANGUAGES = {
    "en": {"flag": "gb", "name": "English"},
    "pt": {"flag": "pt", "name": "Portuguese"},
    "pt_BR": {"flag": "br", "name": "Pt Brazil"},
    "es": {"flag": "es", "name": "Spanish"},
    "fr": {"flag": "fr", "name": "French"},
    "de": {"flag": "de", "name": "German"},
    "zh": {"flag": "cn", "name": "Chinese"},
    "ru": {"flag": "ru", "name": "Russian"},
    "pl": {"flag": "pl", "name": "Polish"},
    "el": {"flag": "gr", "name": "Greek"},
    "ja_JP": {"flag": "jp", "name": "Japanese"},
}

FAB_API_MAX_PAGE_SIZE = 100


def custom_password_validator(password: str) -> None:
    """
    A simplistic example for a password validator
    """
    if len(password) < 8:
        raise PasswordComplexityValidationError("Must have at least 8 characters")


# FAB_PASSWORD_COMPLEXITY_VALIDATOR = custom_password_validator

FAB_PASSWORD_COMPLEXITY_ENABLED = True

# ------------------------------
# GLOBALS FOR GENERAL APP's
# ------------------------------
UPLOAD_FOLDER = basedir + "/app/static/uploads/"
IMG_UPLOAD_FOLDER = basedir + "/app/static/uploads/"
IMG_UPLOAD_URL = "/static/uploads/"
AUTH_TYPE = AUTH_DB
# AUTH_LDAP_SERVER = "ldap://dc.domain.net"
AUTH_ROLE_ADMIN = "Admin"
AUTH_ROLE_PUBLIC = "Public"
APP_NAME = "F.A.B. Example"
APP_THEME = ""  # default
# APP_THEME = "cerulean.css"      # COOL
# APP_THEME = "amelia.css"
# APP_THEME = "cosmo.css"
# APP_THEME = "cyborg.css"       # COOL
# APP_THEME = "flatly.css"
# APP_THEME = "journal.css"
# APP_THEME = "readable.css"
# APP_THEME = "simplex.css"
# APP_THEME = "slate.css"          # COOL
# APP_THEME = "spacelab.css"      # NICE
# APP_THEME = "united.css"
# APP_THEME = "darkly.css"
# APP_THEME = "lumen.css"
# APP_THEME = "paper.css"
# APP_THEME = "sandstone.css"
# APP_THEME = "solar.css"
# APP_THEME = "superhero.css"



import os
basedir = os.path.abspath(os.path.dirname(__file__))

CSRF_ENABLED = True
SECRET_KEY = '\2\1thisismyscretkey\1\2\e\y\y\h'

OPENID_PROVIDERS = [
    { 'name': 'Google', 'url': 'https://www.google.com/accounts/o8/id' },
    { 'name': 'Yahoo', 'url': 'https://me.yahoo.com' },
    { 'name': 'AOL', 'url': 'http://openid.aol.com/<username>' },
    { 'name': 'Flickr', 'url': 'http://www.flickr.com/<username>' },
    { 'name': 'MyOpenID', 'url': 'https://www.myopenid.com' }]

SQLALCHEMY_DATABASE_URI = 'sqlite:///' + os.path.join(basedir, 'app.db')
#SQLALCHEMY_DATABASE_URI = 'mysql://myapp@localhost/myapp'
SQLALCHEMY_MIGRATE_REPO = os.path.join(basedir, 'db_repository')

# administrator list
ADMINS = ['you@example.com']

# pagination
POSTS_PER_PAGE = 3
MAX_SEARCH_RESULTS = 50

BABEL_DEFAULT_LOCALE = 'en'

LANGUAGES = {
    'en': {'flag':'gb', 'name':'English'},
    'pt': {'flag':'pt', 'name':'Portugal'}
}



#------------------------------
# GLOBALS FOR GENERAL APP's
#------------------------------
UPLOAD_FOLDER = basedir + '/app/static/uploads/'
IMG_UPLOAD_FOLDER = basedir + '/app/static/uploads/'
IMG_UPLOAD_URL = '/static/uploads/'
AUTH_TYPE = 1
AUTH_ROLE_ADMIN = 'Admin'
AUTH_ROLE_PUBLIC = 'Public'
APP_NAME = "My App 0.2"
APP_THEME = ""                  # default
#APP_THEME = "cerulean.css"      # COOL
#APP_THEME = "amelia.css"
#APP_THEME = "cosmo.css"
#APP_THEME = "cyborg.css"       # COOL
#APP_THEME = "flatly.css"
#APP_THEME = "journal.css"
#APP_THEME = "readable.css"
#APP_THEME = "simplex.css"
#APP_THEME = "slate.css"          # COOL
#APP_THEME = "spacelab.css"      # NICE
#APP_THEME = "united.css"



from datetime import datetime
import json
import logging
from typing import Dict, List, Optional, Tuple, Union
import uuid

from sqlalchemy import and_, func, literal, update
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import contains_eager
from sqlalchemy.orm.exc import MultipleResultsFound
from werkzeug.security import generate_password_hash

from .apis import (
    GroupApi,
    PermissionApi,
    PermissionViewMenuApi,
    RoleApi,
    UserApi,
    ViewMenuApi,
)
from .models import (
    assoc_permissionview_role,
    Group,
    Permission,
    PermissionView,
    RegisterUser,
    Role,
    User,
    ViewMenu,
)
from ..manager import BaseSecurityManager
from ... import const as c
from ...models.sqla import Base
from ...models.sqla.interface import SQLAInterface

log = logging.getLogger(__name__)


class SecurityManager(BaseSecurityManager):
    """
    Responsible for authentication, registering security views,
    role and permission auto management

    If you want to change anything just inherit and override, then
    pass your own security manager to AppBuilder.
    """

    user_model = User
    """ Override to set your own User Model """
    role_model = Role
    """ Override to set your own Role Model """
    group_model = Group
    permission_model = Permission
    viewmenu_model = ViewMenu
    permissionview_model = PermissionView
    registeruser_model = RegisterUser

    # APIs
    permission_api = PermissionApi
    role_api = RoleApi
    user_api = UserApi
    view_menu_api = ViewMenuApi
    permission_view_menu_api = PermissionViewMenuApi
    group_api = GroupApi

    def __init__(self, appbuilder):
        """
        SecurityManager contructor
        param appbuilder:
            F.A.B AppBuilder main object
        """
        super(SecurityManager, self).__init__(appbuilder)
        user_datamodel = SQLAInterface(self.user_model)
        if self.auth_type == c.AUTH_DB:
            self.userdbmodelview.datamodel = user_datamodel
        elif self.auth_type == c.AUTH_LDAP:
            self.userldapmodelview.datamodel = user_datamodel
        elif self.auth_type == c.AUTH_OID:
            self.useroidmodelview.datamodel = user_datamodel
        elif self.auth_type == c.AUTH_OAUTH:
            self.useroauthmodelview.datamodel = user_datamodel
        elif self.auth_type == c.AUTH_REMOTE_USER:
            self.userremoteusermodelview.datamodel = user_datamodel

        if self.userstatschartview:
            self.userstatschartview.datamodel = user_datamodel
        if self.auth_user_registration:
            self.registerusermodelview.datamodel = SQLAInterface(
                self.registeruser_model
            )

        self.rolemodelview.datamodel = SQLAInterface(self.role_model)
        self.groupmodelview.datamodel = SQLAInterface(self.group_model)
        self.permissionmodelview.datamodel = SQLAInterface(self.permission_model)
        self.viewmenumodelview.datamodel = SQLAInterface(self.viewmenu_model)
        self.permissionviewmodelview.datamodel = SQLAInterface(
            self.permissionview_model
        )
        self.create_db()

    @property
    def get_session(self):
        return self.appbuilder.get_session

    def register_views(self):
        super(SecurityManager, self).register_views()

        if self.appbuilder.app.config.get("FAB_ADD_SECURITY_API", False):
            self.appbuilder.add_api(self.permission_api)
            self.appbuilder.add_api(self.role_api)
            self.appbuilder.add_api(self.user_api)
            self.appbuilder.add_api(self.view_menu_api)
            self.appbuilder.add_api(self.permission_view_menu_api)
            self.appbuilder.add_api(self.group_api)

    def create_db(self):
        try:
            engine = self.get_session.get_bind(mapper=None, clause=None)
            inspector = Inspector.from_engine(engine)
            existing_tables = inspector.get_table_names()
            if "ab_user" not in existing_tables or "ab_group" not in existing_tables:
                log.info(c.LOGMSG_INF_SEC_NO_DB)
                Base.metadata.create_all(engine)
                log.info(c.LOGMSG_INF_SEC_ADD_DB)
            super(SecurityManager, self).create_db()
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_CREATE_DB, e)
            exit(1)

    def find_register_user(self, registration_hash):
        return (
            self.get_session.query(self.registeruser_model)
            .filter(self.registeruser_model.registration_hash == registration_hash)
            .scalar()
        )

    def add_register_user(
        self, username, first_name, last_name, email, password="", hashed_password=""
    ):
        """
        Add a registration request for the user.

        :rtype : RegisterUser
        """
        register_user = self.registeruser_model()
        register_user.username = username
        register_user.email = email
        register_user.first_name = first_name
        register_user.last_name = last_name
        if hashed_password:
            register_user.password = hashed_password
        else:
            register_user.password = generate_password_hash(
                password=password,
                method=self.appbuilder.get_app.config.get(
                    "FAB_PASSWORD_HASH_METHOD", "scrypt"
                ),
                salt_length=self.appbuilder.get_app.config.get(
                    "FAB_PASSWORD_HASH_SALT_LENGTH", 16
                ),
            )
        register_user.registration_hash = str(uuid.uuid1())
        try:
            self.get_session.add(register_user)
            self.get_session.commit()
            return register_user
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_ADD_REGISTER_USER, e)
            self.appbuilder.get_session.rollback()
            return None

    def del_register_user(self, register_user):
        """
        Deletes registration object from database

        :param register_user: RegisterUser object to delete
        """
        try:
            self.get_session.delete(register_user)
            self.get_session.commit()
            return True
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_DEL_REGISTER_USER, e)
            self.get_session.rollback()
            return False

    def find_user(self, username=None, email=None):
        """
        Finds user by username or email
        """
        if username:
            try:
                if self.auth_username_ci:
                    return (
                        self.get_session.query(self.user_model)
                        .filter(
                            func.lower(self.user_model.username) == func.lower(username)
                        )
                        .one_or_none()
                    )
                else:
                    return (
                        self.get_session.query(self.user_model)
                        .filter(self.user_model.username == username)
                        .one_or_none()
                    )
            except MultipleResultsFound:
                log.error("Multiple results found for user %s", username)
                return None
        elif email:
            try:
                return (
                    self.get_session.query(self.user_model)
                    .filter_by(email=email)
                    .one_or_none()
                )
            except MultipleResultsFound:
                log.error("Multiple results found for user with email %s", email)
                return None

    def get_all_users(self):
        return self.get_session.query(self.user_model).all()

    def add_user(
        self,
        username: str,
        first_name: str,
        last_name: str,
        email: str,
        role: Union[List[Role], Role, None] = None,
        password: str = "",
        hashed_password: str = "",
        groups: Optional[List[Group]] = None,
    ):
        """
        Generic function to create user
        """
        roles = []
        if role:
            roles = role if isinstance(role, list) else [role]

        try:
            user = self.user_model()
            user.first_name = first_name
            user.last_name = last_name
            user.username = username
            user.email = email
            user.active = True
            user.roles = roles
            user.groups = groups or []
            if hashed_password:
                user.password = hashed_password
            else:
                user.password = generate_password_hash(
                    password=password,
                    method=self.appbuilder.get_app.config.get(
                        "FAB_PASSWORD_HASH_METHOD", "scrypt"
                    ),
                    salt_length=self.appbuilder.get_app.config.get(
                        "FAB_PASSWORD_HASH_SALT_LENGTH", 16
                    ),
                )
            self.get_session.add(user)
            self.get_session.commit()
            log.info(c.LOGMSG_INF_SEC_ADD_USER, username)
            return user
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_ADD_USER, e)
            self.get_session.rollback()
            return False

    def count_users(self):
        return self.get_session.query(func.count(self.user_model.id)).scalar()

    def update_user(self, user):
        try:
            self.get_session.merge(user)
            self.get_session.commit()
            log.info(c.LOGMSG_INF_SEC_UPD_USER, user)
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_UPD_USER, e)
            self.get_session.rollback()
            return False

    def get_user_by_id(self, pk):
        return self.get_session.query(self.user_model).get(pk)

    def get_first_user(self) -> "User":
        return self.get_session.query(self.user_model).first()

    def noop_user_update(self, user: "User") -> None:
        stmt = (
            update(self.user_model)
            .where(self.user_model.id == user.id)
            .values(login_count=user.login_count)
        )
        self.get_session.execute(stmt)
        self.get_session.commit()

    """
    -----------------------
     PERMISSION MANAGEMENT
    -----------------------
    """

    def add_role(
        self, name: str, permissions: Optional[List[PermissionView]] = None
    ) -> Optional[Role]:
        if not permissions:
            permissions = []

        role = self.find_role(name)
        if role is None:
            try:
                role = self.role_model()
                role.name = name
                role.permissions = permissions
                self.get_session.add(role)
                self.get_session.commit()
                log.info(c.LOGMSG_INF_SEC_ADD_ROLE, name)
                return role
            except Exception as e:
                log.error(c.LOGMSG_ERR_SEC_ADD_ROLE, e)
                self.get_session.rollback()
        return role

    def update_role(self, pk, name: str) -> Optional[Role]:
        role = self.get_session.query(self.role_model).get(pk)
        if not role:
            return
        try:
            role.name = name
            self.get_session.merge(role)
            self.get_session.commit()
            log.info(c.LOGMSG_INF_SEC_UPD_ROLE, role)
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_UPD_ROLE, e)
            self.get_session.rollback()
            return

    def find_role(self, name):
        return (
            self.get_session.query(self.role_model).filter_by(name=name).one_or_none()
        )

    def get_all_roles(self):
        return self.get_session.query(self.role_model).all()

    def get_public_role(self):
        return (
            self.get_session.query(self.role_model)
            .filter_by(name=self.auth_role_public)
            .one_or_none()
        )

    def find_group(self, name: str) -> Group:
        return (
            self.get_session.query(self.group_model).filter_by(name=name).one_or_none()
        )

    def add_group(
        self,
        name: str,
        label: str,
        description: str,
        roles: Optional[List[Role]] = None,
        users: Optional[List[User]] = None,
    ) -> Optional[Group]:
        group = self.find_group(name)
        if group is not None:
            return group
        try:
            group = self.group_model()
            group.name = name
            group.label = label
            group.description = description
            group.roles = roles or []
            group.users = users or []

            self.get_session.add(group)
            self.get_session.commit()
            log.info(c.LOGMSG_INF_SEC_ADD_ROLE, name)
            return group
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_ADD_GROUP, e)
            self.get_session.rollback()

    def get_public_permissions(self):
        role = self.get_public_role()
        if role:
            return role.permissions
        return []

    def find_permission(self, name):
        """
        Finds and returns a Permission by name
        """
        return (
            self.get_session.query(self.permission_model)
            .filter_by(name=name)
            .one_or_none()
        )

    def exist_permission_on_roles(
        self, view_name: str, permission_name: str, role_ids: List[int]
    ) -> bool:
        """
            Method to efficiently check if a certain permission exists
            on a list of role id's. This is used by `has_access`

        :param view_name: The view's name to check if exists on one of the roles
        :param permission_name: The permission name to check if exists
        :param role_ids: a list of Role ids
        :return: Boolean
        """
        q = (
            self.appbuilder.get_session.query(self.permissionview_model)
            .join(
                assoc_permissionview_role,
                and_(
                    (
                        self.permissionview_model.id
                        == assoc_permissionview_role.c.permission_view_id
                    )
                ),
            )
            .join(self.role_model)
            .join(self.permission_model)
            .join(self.viewmenu_model)
            .filter(
                self.viewmenu_model.name == view_name,
                self.permission_model.name == permission_name,
                self.role_model.id.in_(role_ids),
            )
            .exists()
        )
        # Special case for MSSQL/Oracle (works on PG and MySQL > 8)
        if self.appbuilder.get_session.bind.dialect.name in ("mssql", "oracle"):
            return self.appbuilder.get_session.query(literal(True)).filter(q).scalar()
        return self.appbuilder.get_session.query(q).scalar()

    def find_roles_permission_view_menus(
        self, permission_name: str, role_ids: List[int]
    ):
        return (
            self.appbuilder.get_session.query(self.permissionview_model)
            .join(
                assoc_permissionview_role,
                and_(
                    (
                        self.permissionview_model.id
                        == assoc_permissionview_role.c.permission_view_id
                    )
                ),
            )
            .join(self.role_model)
            .join(self.permission_model)
            .join(self.viewmenu_model)
            .filter(
                self.permission_model.name == permission_name,
                self.role_model.id.in_(role_ids),
            )
        ).all()

    def get_user_roles_permissions(self, user) -> Dict[str, List[Tuple[str, str]]]:
        """
        Utility method for fetching all roles and permissions for a specific user.
        Example of the returned data:
        ```
        {
            'Admin': [
                ('can_this_form_get', 'ResetPasswordView'),
                ('can_this_form_post', 'ResetPasswordView'),
                ...
            ]
             'EmptyRole': [],
        }
        ```
        """
        if not user.roles and not user.groups:
            raise AttributeError("User object does not have roles or groups")

        result: Dict[str, List[Tuple[str, str]]] = {}
        db_roles_ids = []
        roles = self.get_user_roles(user)
        for role in roles:
            # Make sure all db roles are included on the result
            result[role.name] = []
            if role.name in self.builtin_roles:
                for permission in self.builtin_roles[role.name]:
                    result[role.name].append((permission[1], permission[0]))
            else:
                db_roles_ids.append(role.id)

        permission_views = (
            self.appbuilder.get_session.query(PermissionView)
            .join(Permission)
            .join(ViewMenu)
            .join(PermissionView.role)
            .filter(Role.id.in_(db_roles_ids))
            .options(contains_eager(PermissionView.permission))
            .options(contains_eager(PermissionView.view_menu))
            .options(contains_eager(PermissionView.role))
        ).all()

        for permission_view in permission_views:
            for role_item in permission_view.role:
                if role_item.name in result:
                    result[role_item.name].append(
                        (
                            permission_view.permission.name,
                            permission_view.view_menu.name,
                        )
                    )
        return result

    def get_db_role_permissions(self, role_id: int) -> List[PermissionView]:
        """
        Get all DB permissions from a role (one single query)
        """
        return (
            self.appbuilder.get_session.query(PermissionView)
            .join(Permission)
            .join(ViewMenu)
            .join(PermissionView.role)
            .filter(Role.id == role_id)
            .options(contains_eager(PermissionView.permission))
            .options(contains_eager(PermissionView.view_menu))
            .all()
        )

    def add_permission(self, name):
        """
        Adds a permission to the backend, model permission

        :param name:
            name of the permission: 'can_add','can_edit' etc...
        """
        perm = self.find_permission(name)
        if perm is None:
            try:
                perm = self.permission_model()
                perm.name = name
                self.get_session.add(perm)
                self.get_session.commit()
                return perm
            except Exception as e:
                log.error(c.LOGMSG_ERR_SEC_ADD_PERMISSION, e)
                self.get_session.rollback()
        return perm

    def del_permission(self, name: str) -> bool:
        """
        Deletes a permission from the backend, model permission

        :param name:
            name of the permission: 'can_add','can_edit' etc...
        """
        perm = self.find_permission(name)
        if not perm:
            log.warning(c.LOGMSG_WAR_SEC_DEL_PERMISSION, name)
            return False
        try:
            pvms = (
                self.get_session.query(self.permissionview_model)
                .filter(self.permissionview_model.permission == perm)
                .all()
            )
            if pvms:
                log.warning(c.LOGMSG_WAR_SEC_DEL_PERM_PVM, perm, pvms)
                return False
            self.get_session.delete(perm)
            self.get_session.commit()
            return True
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_DEL_PERMISSION, e)
            self.get_session.rollback()
            return False

    """
    ----------------------
     PRIMITIVES VIEW MENU
    ----------------------
    """

    def find_view_menu(self, name):
        """
        Finds and returns a ViewMenu by name
        """
        return (
            self.get_session.query(self.viewmenu_model)
            .filter_by(name=name)
            .one_or_none()
        )

    def get_all_view_menu(self):
        return self.get_session.query(self.viewmenu_model).all()

    def add_view_menu(self, name):
        """
        Adds a view or menu to the backend, model view_menu
        param name:
            name of the view menu to add
        """
        view_menu = self.find_view_menu(name)
        if view_menu is None:
            try:
                view_menu = self.viewmenu_model()
                view_menu.name = name
                self.get_session.add(view_menu)
                self.get_session.commit()
                return view_menu
            except Exception as e:
                log.error(c.LOGMSG_ERR_SEC_ADD_VIEWMENU, e)
                self.get_session.rollback()
        return view_menu

    def del_view_menu(self, name: str) -> bool:
        """
        Deletes a ViewMenu from the backend

        :param name:
            name of the ViewMenu
        """
        view_menu = self.find_view_menu(name)
        if not view_menu:
            log.warning(c.LOGMSG_WAR_SEC_DEL_VIEWMENU, name)
            return False
        try:
            pvms = (
                self.get_session.query(self.permissionview_model)
                .filter(self.permissionview_model.view_menu == view_menu)
                .all()
            )
            if pvms:
                log.warning(c.LOGMSG_WAR_SEC_DEL_VIEWMENU_PVM, view_menu, pvms)
                return False
            self.get_session.delete(view_menu)
            self.get_session.commit()
            return True
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_DEL_PERMISSION, e)
            self.get_session.rollback()
            return False

    """
    ----------------------
     PERMISSION VIEW MENU
    ----------------------
    """

    def find_permission_view_menu(self, permission_name, view_menu_name):
        """
        Finds and returns a PermissionView by names
        """
        permission = self.find_permission(permission_name)
        view_menu = self.find_view_menu(view_menu_name)
        if permission and view_menu:
            return (
                self.get_session.query(self.permissionview_model)
                .filter_by(permission=permission, view_menu=view_menu)
                .one_or_none()
            )

    def find_permissions_view_menu(self, view_menu):
        """
        Finds all permissions from ViewMenu, returns list of PermissionView

        :param view_menu: ViewMenu object
        :return: list of PermissionView objects
        """
        return (
            self.get_session.query(self.permissionview_model)
            .filter_by(view_menu_id=view_menu.id)
            .all()
        )

    def add_permission_view_menu(self, permission_name, view_menu_name):
        """
        Adds a permission on a view or menu to the backend

        :param permission_name:
            name of the permission to add: 'can_add','can_edit' etc...
        :param view_menu_name:
            name of the view menu to add
        """
        if not (permission_name and view_menu_name):
            return None
        pv = self.find_permission_view_menu(permission_name, view_menu_name)
        if pv:
            return pv
        vm = self.add_view_menu(view_menu_name)
        perm = self.add_permission(permission_name)
        pv = self.permissionview_model()
        pv.view_menu, pv.permission = vm, perm
        try:
            self.get_session.add(pv)
            self.get_session.commit()
            log.info(c.LOGMSG_INF_SEC_ADD_PERMVIEW, pv)
            return pv
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_ADD_PERMVIEW, e)
            self.get_session.rollback()

    def del_permission_view_menu(self, permission_name, view_menu_name, cascade=True):
        if not (permission_name and view_menu_name):
            return
        pv = self.find_permission_view_menu(permission_name, view_menu_name)
        if not pv:
            return
        roles_pvs = (
            self.get_session.query(self.role_model)
            .filter(self.role_model.permissions.contains(pv))
            .first()
        )
        if roles_pvs:
            log.warning(
                c.LOGMSG_WAR_SEC_DEL_PERMVIEW,
                view_menu_name,
                permission_name,
                roles_pvs,
            )
            return
        try:
            # delete permission on view
            self.get_session.delete(pv)
            self.get_session.commit()
            # if no more permission on permission view, delete permission
            if not cascade:
                return
            if (
                not self.get_session.query(self.permissionview_model)
                .filter_by(permission=pv.permission)
                .all()
            ):
                self.del_permission(pv.permission.name)
            log.info(c.LOGMSG_INF_SEC_DEL_PERMVIEW, permission_name, view_menu_name)
        except Exception as e:
            log.error(c.LOGMSG_ERR_SEC_DEL_PERMVIEW, e)
            self.get_session.rollback()

    def exist_permission_on_views(self, lst, item):
        for i in lst:
            if i.permission and i.permission.name == item:
                return True
        return False

    def exist_permission_on_view(self, lst, permission, view_menu):
        for i in lst:
            if i.permission.name == permission and i.view_menu.name == view_menu:
                return True
        return False

    def add_permission_role(self, role, perm_view):
        """
        Add permission-ViewMenu object to Role

        :param role:
            The role object
        :param perm_view:
            The PermissionViewMenu object
        """
        if perm_view and perm_view not in role.permissions:
            try:
                role.permissions.append(perm_view)
                self.get_session.merge(role)
                self.get_session.commit()
                log.info(c.LOGMSG_INF_SEC_ADD_PERMROLE, perm_view, role.name)
            except Exception as e:
                log.error(c.LOGMSG_ERR_SEC_ADD_PERMROLE, e)
                self.get_session.rollback()

    def del_permission_role(self, role, perm_view):
        """
        Remove permission-ViewMenu object to Role

        :param role:
            The role object
        :param perm_view:
            The PermissionViewMenu object
        """
        if perm_view in role.permissions:
            try:
                role.permissions.remove(perm_view)
                self.get_session.merge(role)
                self.get_session.commit()
                log.info(c.LOGMSG_INF_SEC_DEL_PERMROLE, perm_view, role.name)
            except Exception as e:
                log.error(c.LOGMSG_ERR_SEC_DEL_PERMROLE, e)
                self.get_session.rollback()

    def export_roles(
        self, path: Optional[str] = None, indent: Optional[Union[int, str]] = None
    ) -> None:
        """Exports roles to JSON file."""
        timestamp = datetime.now().strftime("%Y%m%dT%H%M%S")
        filename = path or f"roles_export_{timestamp}.json"

        serialized_roles = []

        for role in self.get_all_roles():
            serialized_role = {"name": role.name, "permissions": []}
            for pvm in role.permissions:
                permission = pvm.permission
                view_menu = pvm.view_menu
                permission_view_menu = {
                    "permission": {"name": permission.name},
                    "view_menu": {"name": view_menu.name},
                }
                serialized_role["permissions"].append(permission_view_menu)
            serialized_roles.append(serialized_role)

        with open(filename, "w") as fd:
            fd.write(json.dumps(serialized_roles, indent=indent))

    def import_roles(self, path: str) -> None:
        """Imports roles from JSON file."""

        session = self.get_session()

        with open(path, "r") as fd:
            roles_json = json.loads(fd.read())

        roles = []

        for role_kwargs in roles_json:
            role = self.add_role(role_kwargs["name"])
            permission_view_menus = [
                self.add_permission_view_menu(
                    permission_name=pvm_kwargs["permission"]["name"],
                    view_menu_name=pvm_kwargs["view_menu"]["name"],
                )
                for pvm_kwargs in role_kwargs["permissions"]
            ]

            for permission_view_menu in permission_view_menus:
                if permission_view_menu not in role.permissions:
                    role.permissions.append(permission_view_menu)
            roles.append(role)

        session.add_all(roles)
        session.commit()



from flask_appbuilder import ModelView
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.views import MasterDetailView
from flask_babel import lazy_gettext as _

from . import appbuilder, db
from .models import Contact, ContactGroup, Gender


def fill_gender():
    try:
        db.session.add(Gender(name="Male"))
        db.session.add(Gender(name="Female"))
        db.session.commit()
    except Exception:
        db.session.rollback()


class ContactGeneralView(ModelView):
    datamodel = SQLAInterface(Contact)

    label_columns = {"contact_group": "Contacts Group"}
    list_columns = ["name", "personal_phone", "contact_group"]

    base_order = ("name", "asc")

    show_fieldsets = [
        ("Summary", {"fields": ["name", "gender", "contact_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "address",
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                ],
                "expanded": False,
            },
        ),
    ]

    add_fieldsets = [
        ("Summary", {"fields": ["name", "gender", "contact_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "address",
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                ],
                "expanded": False,
            },
        ),
    ]

    edit_fieldsets = [
        ("Summary", {"fields": ["name", "gender", "contact_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "address",
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                ],
                "expanded": False,
            },
        ),
    ]


class GroupMasterView(MasterDetailView):
    datamodel = SQLAInterface(ContactGroup)
    related_views = [ContactGeneralView]


class GroupGeneralView(ModelView):
    datamodel = SQLAInterface(ContactGroup)
    related_views = [ContactGeneralView]


fixed_translations_import = [
    _("List Groups"),
    _("Manage Groups"),
    _("List Contacts"),
    _("Contacts Chart"),
    _("Contacts Birth Chart"),
]

db.create_all()
fill_gender()
appbuilder.add_view(
    GroupMasterView, "List Groups", icon="fa-folder-open-o", category="Contacts"
)
appbuilder.add_separator("Contacts")
appbuilder.add_view(
    GroupGeneralView, "Manage Groups", icon="fa-folder-open-o", category="Contacts"
)
appbuilder.add_view(
    ContactGeneralView, "List Contacts", icon="fa-envelope", category="Contacts"
)



from flask_appbuilder import ModelRestApi
from flask_appbuilder.api import BaseApi, expose
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.models.filters import BaseFilter
from sqlalchemy import or_

from . import appbuilder, db
from .models import Contact, ContactGroup, ModelOMParent
from marshmallow import fields, Schema


db.create_all()


class GreetingsResponseSchema(Schema):
    message = fields.String()


class GreetingApi(BaseApi):
    resource_name = "greeting"
    openapi_spec_component_schemas = (GreetingsResponseSchema,)

    openapi_spec_methods = {
        "greeting": {"get": {"description": "Override description"}}
    }

    @expose("/")
    def greeting(self):
        """Send a greeting
        ---
        get:
          responses:
            200:
              description: Greet the user
              content:
                application/json:
                  schema:
                    type: object
                    $ref: '#/components/schemas/GreetingsResponseSchema'
        """
        return self.response(200, message="Hello")


appbuilder.add_api(GreetingApi)


class CustomFilter(BaseFilter):
    name = "Custom Filter"
    arg_name = "opr"

    def apply(self, query, value):
        return query.filter(
            or_(Contact.name.like(value + "%"), Contact.address.like(value + "%"))
        )


class ContactModelApi(ModelRestApi):
    resource_name = "contact"
    datamodel = SQLAInterface(Contact)
    allow_browser_login = True

    search_filters = {"name": [CustomFilter]}
    openapi_spec_methods = {
        "get_list": {"get": {"description": "Get all contacts, filter and pagination"}}
    }


appbuilder.add_api(ContactModelApi)


class GroupModelApi(ModelRestApi):
    resource_name = "group"
    datamodel = SQLAInterface(ContactGroup)
    allow_browser_login = True


appbuilder.add_api(GroupModelApi)


class ModelOMParentApi(ModelRestApi):
    allow_browser_login = True
    datamodel = SQLAInterface(ModelOMParent)


appbuilder.add_api(ModelOMParentApi)



import logging
from flask import Flask
from flask_appbuilder import AppBuilder
from flask_sqlalchemy import SQLAlchemy
from flask_appbuilder import SQLA
from sqlalchemy import event

logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)

app = Flask(__name__)
app.config.from_object("config")
db = SQLA(app)
session = db.session
appbuilder = AppBuilder(app, session)

"""
Only include this for SQLLite constraints

@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()
"""
from app.models import *


db.create_all()

from app import views



import datetime

from flask import g
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    Sequence,
    String,
    Table,
    UniqueConstraint,
)
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import backref, relationship

from ... import Model
from ..._compat import as_unicode

_dont_audit = False


class Permission(Model):
    __tablename__ = "ab_permission"
    id = Column(Integer, Sequence("ab_permission_id_seq"), primary_key=True)
    name = Column(String(100), unique=True, nullable=False)

    def __repr__(self):
        return self.name


class ViewMenu(Model):
    __tablename__ = "ab_view_menu"
    id = Column(Integer, Sequence("ab_view_menu_id_seq"), primary_key=True)
    name = Column(String(250), unique=True, nullable=False)

    def __eq__(self, other):
        return (isinstance(other, self.__class__)) and (self.name == other.name)

    def __neq__(self, other):
        return self.name != other.name

    def __repr__(self):
        return self.name


assoc_permissionview_role = Table(
    "ab_permission_view_role",
    Model.metadata,
    Column("id", Integer, Sequence("ab_permission_view_role_id_seq"), primary_key=True),
    Column(
        "permission_view_id",
        Integer,
        ForeignKey("ab_permission_view.id", ondelete="CASCADE"),
    ),
    Column("role_id", Integer, ForeignKey("ab_role.id", ondelete="CASCADE")),
    UniqueConstraint("permission_view_id", "role_id"),
    Index("idx_permission_view_id", "permission_view_id"),
    Index("idx_role_id", "role_id"),
)


class Role(Model):
    __tablename__ = "ab_role"

    id = Column(Integer, Sequence("ab_role_id_seq"), primary_key=True)
    name = Column(String(64), unique=True, nullable=False)
    permissions = relationship(
        "PermissionView",
        secondary=assoc_permissionview_role,
        backref="role",
        passive_deletes=True,
    )

    def __repr__(self):
        return self.name


class PermissionView(Model):
    __tablename__ = "ab_permission_view"
    __table_args__ = (
        UniqueConstraint("permission_id", "view_menu_id"),
        Index("idx_permission_id", "permission_id"),
        Index("idx_view_menu_id", "view_menu_id"),
    )
    id = Column(Integer, Sequence("ab_permission_view_id_seq"), primary_key=True)
    permission_id = Column(Integer, ForeignKey("ab_permission.id"))
    permission = relationship("Permission", lazy="joined")
    view_menu_id = Column(Integer, ForeignKey("ab_view_menu.id"))
    view_menu = relationship("ViewMenu", lazy="joined")

    def __repr__(self):
        return str(self.permission).replace("_", " ") + " on " + str(self.view_menu)


assoc_user_role = Table(
    "ab_user_role",
    Model.metadata,
    Column("id", Integer, Sequence("ab_user_role_id_seq"), primary_key=True),
    Column("user_id", Integer, ForeignKey("ab_user.id", ondelete="CASCADE")),
    Column("role_id", Integer, ForeignKey("ab_role.id", ondelete="CASCADE")),
    UniqueConstraint("user_id", "role_id"),
)


class User(Model):
    __tablename__ = "ab_user"
    id = Column(Integer, Sequence("ab_user_id_seq"), primary_key=True)
    first_name = Column(String(64), nullable=False)
    last_name = Column(String(64), nullable=False)
    username = Column(String(64), unique=True, nullable=False)
    password = Column(String(256))
    active = Column(Boolean)
    email = Column(String(320), unique=True, nullable=False)
    last_login = Column(DateTime)
    login_count = Column(Integer)
    fail_login_count = Column(Integer)
    roles = relationship(
        "Role", secondary=assoc_user_role, backref="user", passive_deletes=True
    )
    created_on = Column(
        DateTime, default=lambda: datetime.datetime.now(), nullable=True
    )
    changed_on = Column(
        DateTime, default=lambda: datetime.datetime.now(), nullable=True
    )

    @declared_attr
    def created_by_fk(self):
        return Column(
            Integer, ForeignKey("ab_user.id"), default=self.get_user_id, nullable=True
        )

    @declared_attr
    def changed_by_fk(self):
        return Column(
            Integer, ForeignKey("ab_user.id"), default=self.get_user_id, nullable=True
        )

    created_by = relationship(
        "User",
        backref=backref("created", uselist=True),
        remote_side=[id],
        primaryjoin="User.created_by_fk == User.id",
        uselist=False,
    )
    changed_by = relationship(
        "User",
        backref=backref("changed", uselist=True),
        remote_side=[id],
        primaryjoin="User.changed_by_fk == User.id",
        uselist=False,
    )

    @classmethod
    def get_user_id(cls):
        try:
            return g.user.id
        except Exception:
            return None

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return self.active

    @property
    def is_anonymous(self):
        return False

    def get_id(self):
        return as_unicode(self.id)

    def get_full_name(self):
        return "{0} {1}".format(self.first_name, self.last_name)

    def __repr__(self):
        return self.get_full_name()


assoc_user_group = Table(
    "ab_user_group",
    Model.metadata,
    Column("id", Integer, Sequence("ab_user_group_id_seq"), primary_key=True),
    Column("user_id", Integer, ForeignKey("ab_user.id", ondelete="CASCADE")),
    Column("group_id", Integer, ForeignKey("ab_group.id", ondelete="CASCADE")),
    UniqueConstraint("user_id", "group_id"),
    Index("idx_user_id", "user_id"),
    Index("idx_user_group_id", "group_id"),
)


assoc_group_role = Table(
    "ab_group_role",
    Model.metadata,
    Column("id", Integer, Sequence("ab_group_role_id_seq"), primary_key=True),
    Column("group_id", Integer, ForeignKey("ab_group.id", ondelete="CASCADE")),
    Column("role_id", Integer, ForeignKey("ab_role.id", ondelete="CASCADE")),
    UniqueConstraint("group_id", "role_id"),
    Index("idx_group_id", "group_id"),
    Index("idx_group_role_id", "role_id"),
)


class Group(Model):
    __tablename__ = "ab_group"
    id = Column(Integer, Sequence("ab_group_id_seq"), primary_key=True)
    name = Column(String(100), unique=True, nullable=False)
    label = Column(String(150))
    description = Column(String(512))
    users = relationship(
        "User", secondary=assoc_user_group, backref="groups", passive_deletes=True
    )
    roles = relationship(
        "Role", secondary=assoc_group_role, backref="groups", passive_deletes=True
    )

    def __repr__(self):
        return self.name


class RegisterUser(Model):
    __tablename__ = "ab_register_user"
    id = Column(Integer, Sequence("ab_register_user_id_seq"), primary_key=True)
    first_name = Column(String(64), nullable=False)
    last_name = Column(String(64), nullable=False)
    username = Column(String(64), unique=True, nullable=False)
    password = Column(String(256))
    email = Column(String(64), nullable=False)
    registration_date = Column(DateTime, default=datetime.datetime.now, nullable=True)
    registration_hash = Column(String(256))



import logging

from flask import Flask
from flask_appbuilder import AppBuilder
from flask_appbuilder.security.mongoengine.manager import SecurityManager
from flask_mongoengine import MongoEngine


logging.basicConfig(format="%(asctime)s:%(levelname)s:%(name)s:%(message)s")
logging.getLogger().setLevel(logging.DEBUG)

app = Flask(__name__)
app.config.from_object("config")
dbmongo = MongoEngine(app)
appbuilder = AppBuilder(app, security_manager_class=SecurityManager)

from . import models, views  # noqa



from flask_appbuilder.charts.views import GroupByChartView
from flask_appbuilder.models.group import aggregate_count
from flask_appbuilder.models.sqla.interface import SQLAInterface
from flask_appbuilder.views import ModelView

from . import appbuilder, db
from .models import Person, PersonGroup


class PersonModelView(ModelView):
    datamodel = SQLAInterface(Person, db.session)

    list_title = "List Contacts"
    show_title = "Show Contact"
    add_title = "Add Contact"
    edit_title = "Edit Contact"

    # list_widget = ListThumbnail

    label_columns = {
        "person_group_id": "Group",
        "photo_img": "Photo",
        "photo_img_thumbnail": "Photo",
    }
    list_columns = [
        "photo_img_thumbnail",
        "name",
        "personal_celphone",
        "business_celphone",
        "birthday",
        "person_group",
    ]

    show_fieldsets = [
        ("Summary", {"fields": ["photo_img", "name", "address", "person_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                    "personal_email",
                ],
                "expanded": False,
            },
        ),
        (
            "Professional Info",
            {
                "fields": [
                    "business_function",
                    "business_phone",
                    "business_celphone",
                    "business_email",
                ],
                "expanded": False,
            },
        ),
        ("Extra", {"fields": ["notes"], "expanded": False}),
    ]

    add_fieldsets = [
        ("Summary", {"fields": ["name", "photo", "address", "person_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                    "personal_email",
                ],
                "expanded": False,
            },
        ),
        (
            "Professional Info",
            {
                "fields": [
                    "business_function",
                    "business_phone",
                    "business_celphone",
                    "business_email",
                ],
                "expanded": False,
            },
        ),
        ("Extra", {"fields": ["notes"], "expanded": False}),
    ]

    edit_fieldsets = [
        ("Summary", {"fields": ["name", "photo", "address", "person_group"]}),
        (
            "Personal Info",
            {
                "fields": [
                    "birthday",
                    "personal_phone",
                    "personal_celphone",
                    "personal_email",
                ],
                "expanded": False,
            },
        ),
        (
            "Professional Info",
            {
                "fields": [
                    "business_function",
                    "business_phone",
                    "business_celphone",
                    "business_email",
                ],
                "expanded": False,
            },
        ),
        ("Extra", {"fields": ["notes"], "expanded": False}),
    ]


class GroupModelView(ModelView):
    datamodel = SQLAInterface(PersonGroup, db.session)
    related_views = [PersonModelView]

    label_columns = {"phone1": "Phone (1)", "phone2": "Phone (2)", "taxid": "Tax ID"}
    list_columns = ["name", "notes"]


class PersonChartView(GroupByChartView):
    datamodel = SQLAInterface(Person)
    chart_title = "Grouped Persons"
    label_columns = PersonModelView.label_columns
    chart_type = "PieChart"

    definitions = [
        {"group": "person_group", "series": [(aggregate_count, "person_group")]}
    ]


db.create_all()
appbuilder.add_view(
    GroupModelView(), "List Groups", icon="fa-folder-open-o", category="Contacts"
)
appbuilder.add_view(
    PersonModelView(), "List Contacts", icon="fa-envelope", category="Contacts"
)
appbuilder.add_view(
    PersonChartView(), "Contacts Chart", icon="fa-dashboard", category="Contacts"
)
